-----------------------------------------------------------------------------------------------
-- Client Lua Script for Scryer
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "GameLib"
require "ICCommLib"
require "ChatSystemLib"
require "GuildLib"
require "Apollo"

-----------------------------------------------------------------------------------------------
-- Scryer Module Definition
-----------------------------------------------------------------------------------------------
local Scryer = {} 

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local kstrChannelPrefix = "scryer"
local kstrTitle = "Scryer"
local kstrVersion = "0.6"

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function Scryer:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o.strGuild = nil
	o.tGuild = {}
    -- initialize variables here

    return o
end

--init
function Scryer:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

--save addon status
function Scryer:OnSave(eLevel)
	if (eLevel ~= GameLib.CodeEnumAddonSaveLevel.Account) then 
		return 
	end 

	return {
		update = self.bUpdate,
		ally = self.bShowAlly,
		guild = self.bShowGuild,
		enemy = self.bShowEnemy,
		mediumIcon = self.mediumIcon,
		largeIcon = self.largeIcon,
	}
end

--restore addon status
function Scryer:OnRestore(eLevel,tSavedData) 
	if (eLevel ~= GameLib.CodeEnumAddonSaveLevel.Account) then 
		return 
	end 
	
	if tSavedData.update ~= nil then
		self.bUpdate = tSavedData.update 
	end
	if tSavedData.ally ~= nil then
		self.bShowAlly = tSavedData.ally 
	end
	if tSavedData.guild ~= nil then
		self.bShowGuild = tSavedData.guild 
	end
	if tSavedData.enemy ~= nil then
		self.bShowEnemy = tSavedData.enemy 
	end
	if tSavedData.mediumIcon ~= nil then
		self.mediumIcon = tSavedData.mediumIcon
	end
	if tSavedData.largeIcon ~= nil then
		self.largeIcon = tSavedData.largeIcon
	end

end 

-----------------------------------------------------------------------------------------------
-- Scryer OnLoad
-----------------------------------------------------------------------------------------------
function Scryer:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("Scryer.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- Scryer OnDocLoaded
-----------------------------------------------------------------------------------------------
function Scryer:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.wndMain = Apollo.LoadForm(self.xmlDoc, "ScryerForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
		self.wndMain:Show(false, true)
		self.lastPos = nil
		self.msgChannel = nil
		self.strGuild = nil
		self.tGuild = {}
		self.strZoneName = "Unknown"
		self.iContId = nil
		-- if the xmlDoc is no longer needed, you should set it to nil
		-- self.xmlDoc = nil
		-- load the saved data if it's present or set to defaults
		if self.mediumIcon == nil then
			self.mediumIcon = 0.7
		end
		if self.largeIcon == nil then
			self.largeIcon = 0.3
		end
		if self.bUpdate ~= true then
			Apollo.StopTimer("ScryerUpdateTimer")
		end
		if self.bShowAlly then
			self.wndMain:FindChild("chkAlly"):SetCheck(true)
		end
		if self.bShowGuild then
			self.wndMain:FindChild("chkGuild"):SetCheck(true)
		end
		if self.bShowEnemy then
			self.wndMain:FindChild("chkEnemy"):SetCheck(true)
		end
		self.wndMain:FindChild("Title"):SetText(kstrTitle .. " v" .. kstrVersion)
		self.wndMain:FindChild("txtLarge"):SetText((self.largeIcon*100) .. "%")
		self.wndMain:FindChild("txtMedium"):SetText((self.mediumIcon*100) .. "%")
		-- Register handlers for events, slash commands and timer, etc.
		if self.bShowAlly or self.bShowGuild or self.bShowEnemy then
			Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
			Apollo.RegisterEventHandler("UnitUpdated", "OnUnitCreated", self)
		end
		Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
		Apollo.RegisterSlashCommand("scryer", "OnScryerOn", self)
		--make the timer but turn it off unless it is needed
		Apollo.RegisterTimerHandler("ScryerUpdateTimer", "OnTimer", self)
		Apollo.CreateTimer("ScryerUpdateTimer", 1.0, true)
		Apollo.RegisterTimerHandler("ScryerGarbageTimer", "OnGarbageTimer", self)
		Apollo.CreateTimer("ScryerGarbageTimer", 60.0, true)
		Apollo.RegisterTimerHandler("ScryerUnitTimer", "OnUnitTimer", self)
		Apollo.CreateTimer("ScryerUnitTimer", 1.0, false)
		--Apollo.StopTimer("ScryerUnitTimer")
		--figure out zone for map issues
		Apollo.RegisterEventHandler("VarChange_ZoneName", "OnChangeZoneName", self)
		-- Do additional Addon initialization here
	end
end

-----------------------------------------------------------------------------------------------
-- Scryer Functions
-----------------------------------------------------------------------------------------------
-- on SlashCommand "/scryer" or "/scryer on" or "/scryer off"
function Scryer:OnScryerOn(cmd, args)
	local strMessage = ""
	if args == "on" then
		if self:EnableTracking() then
			strMessage = "Scryer ON"
		else
			strMessage = "Scryer already On"
		end
	elseif args == "off" then
		if self:DisableTracking() then
			strMessage = "Scryer OFF"
		else
			strMessage = "Scryer already Off"
		end
	else
		self.wndMain:Invoke() -- show the window
		return
	end
	--if we've reached this point we're using an argument, inform user
	ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Command, strMessage, "" )
end

-- on timer, do a check and generate message to send if passed
function Scryer:OnTimer()
	if not GameLib.IsCharacterLoaded() then
		return
	end
	if self.bUpdate ~= true then
		return
	end
	self:GenerateMessage()
end

function Scryer:OnUnitTimer()
	self:CheckUnits()
	local len = #self.tGuild
	if len > 10 then
		Apollo.CreateTimer("ScryerUnitTimer", len / 10, false)
	else
		Apollo.CreateTimer("ScryerUnitTimer", 1.0, false)
	end
	Apollo.StartTimer("ScryerUnitTimer")
end

-- on timer, do a garbage handling on players
function Scryer:OnGarbageTimer()
	if GameLib.GetPlayerUnit() ~= nil then
		local strPlayer = GameLib.GetPlayerUnit():GetName()
		if self.tGuild[strPlayer] ~= nil then
			self:RemoveMarkers(strPlayer)
			self.tGuild[strPlayer] = nil
		end
	end

	for idx, temp in pairs(self.tGuild) do
		if not self.tGuild[idx].bUpdated then
			self:RemoveMarkers(idx)
			self.tGuild[idx] = nil
		else
			self.tGuild[idx].bUpdated = false
		end
	end
end

-- find unit if able
function Scryer:CheckUnits()
	if not (self.bShowAlly or self.bShowEnemy or self.bShowGuild) then
		return
	end
	for idx, temp in pairs(self.tGuild) do
		if self.tGuild[idx].nId ~= nil then
			local unit = GameLib.GetUnitById(self.tGuild[idx].nId)
			if unit == nil then
				self:RemoveMarkers(idx)
				self.tGuild[idx] = nil
			else
				if unit:GetHealth() ~= unit:GetMaxHealth() and (unit:GetHealth() < (unit:GetMaxHealth()*self.largeIcon)) and self.tGuild[idx].state ~= 1 then
					self:OnUnitCreated(unit)			
				elseif unit:GetHealth() ~= unit:GetMaxHealth() and (unit:GetHealth() < (unit:GetMaxHealth()*self.mediumIcon)) and self.tGuild[idx].state ~= 2 then
					self:OnUnitCreated(unit)			
				elseif self.tGuild[idx].state ~= 3 then 
					self:OnUnitCreated(unit)	
				elseif self.tGuild[idx].strColor ~= self:GetColor(unit) then
					self:OnUnitCreated(unit)	
				end
			end
		end
	end
end

--build message to send
function Scryer:GenerateMessage()
	if GameLib.GetPlayerUnit() ~= nil then
		--let's not send an update if we haven't moved
		--if not self:UpdatePosition(self.lastPos,GameLib.GetPlayerUnit():GetPosition()) then
		--	return
		--end
		--figure out where we are
		if self.strZoneName == "Unknown" then
			self.strZoneName = self:GetCurrentZoneName()
		end
		self.lastPos = GameLib.GetPlayerUnit():GetPosition()
		--ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Command, "Buttes", "" )
		-- or if we aren't in a guild
		local guild = GameLib.GetPlayerUnit():GetAffiliationName()
		if guild ~= nil then
			--store name, position, and zone
			local t = {}
			t.strPlayer = GameLib.GetPlayerUnit():GetName()
			t.pos = self.lastPos
			t.strZone = self.strZoneName
			t.iContId = self.iContId

			--try to dispatch message
			self:SendMessage(guild, t)
		end
	end
end

--dispatch message to proper channel
function Scryer:SendMessage(channel, tMsg)
	--setup message handler if channel is not set or has changed
	if self.msgChannel == nil or self.strGuild ~= channel then
		self.strGuild = channel
		self.msgChannel = ICCommLib.JoinChannel(kstrChannelPrefix  .. channel, "OnMessageRecv", self)
	end
	self.msgChannel:SendMessage(tMsg)
end

--message received handler
function Scryer:OnMessageRecv(channel, tMsg)
	if not GameLib.IsCharacterLoaded() then
		return
	end
	if tMsg == nil or tMsg.strPlayer == nil or tMsg.pos == nil then
		return
	end
	local strZone = "Unknown"
	if tMsg.strZone ~= nil then
		strZone = tMsg.strZone
	end
	-- do we know the player? if not, add to table
	if self.tGuild[tMsg.strPlayer] == nil then
		self.tGuild[tMsg.strPlayer] = {}
	else
		self:RemoveMarkers(tMsg.strPlayer)
	end
	--are we on the right continent as the player? or looking at it rather?
	--can't seem to get this working trying to integrate with zone map
	--falling back to zone name indicator
	--if self.iContId == nil then
	--	self.iContId = self:GetCurrentContinentId()
	--	if self.iContId == nil then
	--		self.iContId = self:GetMyContinentId()
	--	end
	--end
	--debug
	--local strMessage = tMsg.strPlayer .. self.strZoneName
	--if tMsg.iContId ~= nil then
	--	strMessage = strMessage .. "?" .. tMsg.iContId
	--end
	--ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Command, strMessage, "" )
	--end debug
	--if tMsg.iContId ~= nil and tMsg.iContId ~= self.iContId then
	--	return
	--end
	self.tGuild[tMsg.strPlayer].bUpdated = true
	--add player dot
	self:AddMarkers(nil, tMsg.strPlayer, tMsg.pos, strZone, true, "DispositionGuildmateUnflagged", {bNeverShowOnEdge=true, bFixedSizeSmall=true}, false)
end

--attempt to limit updates
function Scryer:UpdatePosition(posA, posB)
	if posA == nil or posB == nil then
		return true
	end
	local x = posA.x - posB.x
	local y = posA.y - posB.y
	local z = posA.z - posB.z
	local limiter = 1
	if x < limiter and y < limiter and z < limiter then
		return false
	end
	return true
end

--add markers
function Scryer:AddMarkers(unit, strPlayer, pos, strZone, bOnMap, strColor, tFlags, bEnemy)
	if g_wndTheZoneMap == nil or g_wndTheMiniMap == nil then
		return
	end
	--setup marker icon
	local tInfo = {
		strIcon = "IconSprites:Icon_MapNode_Map_GroupMember",
		crObject = ApolloColor.new(strColor),
		strIconEdge = "ClientSprites:MiniMapMarkerTiny",
		crEdge = ApolloColor.new(strColor),
    	}
	if bEnemy then
		tInfo.strIcon = "sprMM_TargetCreature"
	end
	-- add new position dot
	if bOnMap then
		local playerName = strPlayer .. "\n(" .. strZone ..")"
		--ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Command, tColor.r, "" )
		playerName = string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"ff913db7\">%s</T>", strPlayer)
		playerName = playerName .." (" .. strZone .. ")" --String_GetWeaselString(Apollo.GetString("ZoneMap_AppendGroupMemberLabel"), playerName)

		self.tGuild[strPlayer].zmObj = g_wndTheZoneMap:AddObject(GameLib.CodeEnumMapOverlayType.Object, pos, playerName, tInfo, {bNeverShowOnEdge=true, bFixedSizeMedium=true})
	end
	--tInfo.strIcon = "ClientSprites:MiniMapMarkerTiny"
	if unit ~= nil then
		g_wndTheMiniMap:AddUnit(unit, GameLib.CodeEnumMapOverlayType.Unit, tInfo, tFlags, false)
	else
		self.tGuild[strPlayer].mmObj = g_wndTheMiniMap:AddObject(GameLib.CodeEnumMapOverlayType.Object, pos, strPlayer, tInfo, tFlags)
	end
end

--get rid of the markers
function Scryer:RemoveMarkers(strPlayer)
	if g_wndTheZoneMap == nil or g_wndTheMiniMap == nil then
		return
	end
	if self.tGuild[strPlayer] == nil then
		return
	end
	if self.tGuild[strPlayer].zmObj ~= nil then
		g_wndTheZoneMap:RemoveObject(self.tGuild[strPlayer].zmObj)
		self.tGuild[strPlayer].zmObj = nil
	end
	if self.tGuild[strPlayer].mmObj ~= nil then
		g_wndTheMiniMap:RemoveObject(self.tGuild[strPlayer].mmObj)
		self.tGuild[strPlayer].mmObj = nil
	elseif self.tGuild[strPlayer].nId ~= nil then
		local unit = GameLib.GetUnitById(self.tGuild[strPlayer].nId)
		if unit ~= nil then
			g_wndTheMiniMap:RemoveUnit(unit)
		end
	end
end

--turn on tracking if it is off, return result
function Scryer:DisableTracking()
	if self.bUpdate then
		Apollo.StopTimer("ScryerUpdateTimer")
		self.bUpdate = false
		for idx, temp in pairs(self.tGuild) do
			self:RemoveMarkers(idx)
			self.tGuild[idx] = nil
		end
		return true
	end
	return false
end

--turn off tracking if it is on, return result
function Scryer:EnableTracking()
	if not self.bUpdate then
		Apollo.StartTimer("ScryerUpdateTimer")
		self.bUpdate = true
		return true
	end
	return false
end

--zone change handler
function Scryer:OnChangeZoneName(oVar, strNewZone)
	if strNewZone == nil then
		return
	end
	self.strZoneName = self:GetCurrentZoneName()
	self.iContId = self:GetCurrentContinentId()
end

function Scryer:GetCurrentZoneName()
	local currentZone = GameLib.GetCurrentZoneMap()
	if currentZone == nil then
		self.unknown = true
		return "Unknown"
	else
		return currentZone.strName
	end
end

function Scryer:GetCurrentContinentId()
	if g_wndTheZoneMap == nil then
		return nil
	end

	local tCurrentInfo = g_wndTheZoneMap:GetZoneInfo()
	if not tCurrentInfo then
		return nil
	end

	local idCurrentZone = tCurrentInfo.id
	local tCurrentContinent = g_wndTheZoneMap:GetContinentInfo(tCurrentInfo.continentId)
	return tCurrentContinent.id 
end

function Scryer:GetMyContinentId()
	local tHomeZone = GameLib.GetCurrentZoneMap()
	return tHomeZone.continentId
end

function Scryer:OnZoneMapChange(wndHandler, wndControl, strContinent, idZone)
	if idZone ~= nil then
		self.iContId = idZone
	end
end

function Scryer:OnUnitCreated(unit)
	if unit == nil then
		return
	end
	-- make sure the unit created is a character, or this list will
	-- get filled up with a lot of stuff
	if not unit:IsACharacter() then
		return
	end
	local strPlayer = unit:GetName()
	-- do we know the player? if not, add to table
	if self.tGuild[strPlayer] == nil then
		self.tGuild[strPlayer] = {}
	else
		self:RemoveMarkers(strPlayer)
	end

	self.tGuild[strPlayer].bUpdated = true
	self.tGuild[strPlayer].nId = unit:GetId()
	--add player dot
	local strColor = self:GetColor(unit)
	if strColor == "error" then
		return
	end
	local bEnemy = GameLib.GetPlayerUnit():GetFaction() ~= unit:GetFaction()

	if unit:GetHealth() ~= unit:GetMaxHealth() and (unit:GetHealth() < (unit:GetMaxHealth()*self.largeIcon)) then
		self.tGuild[strPlayer].state = 1
		self:AddMarkers(unit, strPlayer, unit:GetPosition(), "", false, strColor, {bNeverShowOnEdge=true, bFixedSizeLarge=true}, bEnemy)
	elseif unit:GetHealth() ~= unit:GetMaxHealth() and (unit:GetHealth() < (unit:GetMaxHealth()*self.mediumIcon)) then
		self.tGuild[strPlayer].state = 2
		self:AddMarkers(unit, strPlayer, unit:GetPosition(), "", false, strColor, {bNeverShowOnEdge=true, bFixedSizeMedium=true}, bEnemy)
	else
		self.tGuild[strPlayer].state = 3
		self:AddMarkers(unit, strPlayer, unit:GetPosition(), "", false, strColor, {bNeverShowOnEdge=true, bFixedSizeSmall=true}, bEnemy)
	end
	self.tGuild[strPlayer].strColor = strColor
end

function Scryer:GetColor(unit)
	if GameLib.GetPlayerUnit() == nil then
		return "error"
	end
	local myFaction = GameLib.GetPlayerUnit():GetFaction() == unit:GetFaction()
	local myGuild = GameLib.GetPlayerUnit():GetAffiliationName() == unit:GetAffiliationName()
	local pvp = unit:IsPvpFlagged()

	if myGuild and self.bShowGuild then
		return "DispositionGuildmateUnflagged"
	end

	--other faction or mine?
	if myFaction and self.bShowAlly then
		if pvp then
			return "DispositionFriendly"
		end
		return "DispositionFriendlyUnflagged"
	elseif not myFaction and self.bShowEnemy then
		if pvp then
			return "DispositionHostile"
		end
		return "DispositionNeutral"
	end

	return "error"
end

function Scryer:OnUnitDestroyed(unit)
	-- make sure the unit destroyed is a character
	if not unit:IsACharacter() then
		return
	end
	-- do we know the player?
	local strPlayer = unit:GetName()
	if self.tGuild[strPlayer] == nil then
		self:RemoveMarkers(strPlayer)
	end
end

function Scryer:Listen()
	-- are we already listening?
	if bShowAlly or bShowGuild or bShowEnemy then
		return
	end

	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitUpdated", "OnUnitCreated", self)
end

function Scryer:Ignore()
	--is anyone listening still?
	if bShowAlly or bShowGuild or bShowEnemy then
		return
	end

	Apollo.RemoveEventHandler("UnitCreated", self)
	Apollo.RemoveEventHandler("UnitUpdated", self)
end
-----------------------------------------------------------------------------------------------
-- ScryerForm Functions
-----------------------------------------------------------------------------------------------
-- when the Cancel button is clicked
function Scryer:OnCancel()
	self.wndMain:Close() -- hide the window
end

-- when the ON button is clicked
function Scryer:OnON( wndHandler, wndControl, eMouseButton )
	local t = self:EnableTracking()
	self.wndMain:Close() -- hide the window
end

-- when the OFF button is clicked
function Scryer:OnOFF( wndHandler, wndControl, eMouseButton )
	local t = self:DisableTracking()
	self.wndMain:Close() -- hide the window
end

function Scryer:OnToggleAlly( wndHandler, wndControl)
	if wndHandler:IsChecked() then
		self:Listen()
		self.bShowAlly = true
	else
		self.bShowAlly = false
		self:Ignore()
	end
end

function Scryer:OnToggleGuild( wndHandler, wndControl)
	if wndHandler:IsChecked() then
		self:Listen()
		self.bShowGuild = true
	else
		self.bShowGuild = false
		self:Ignore()
	end
end

function Scryer:OnToggleEnemy( wndHandler, wndControl)
	if wndHandler:IsChecked() then
		self:Listen()
		self.bShowEnemy = true
	else
		self.bShowEnemy = false
		self:Ignore()
	end
end

function Scryer:OnMediumDown( wndHandler, wndControl, eMouseButton )
	if self.mediumIcon > self.largeIcon then
		self.mediumIcon = self.mediumIcon - 0.05
		self.wndMain:FindChild("txtMedium"):SetText((self.mediumIcon*100) .. "%")
	end
end

function Scryer:OnMediumUp( wndHandler, wndControl, eMouseButton )
	if self.mediumIcon < 100 then
		self.mediumIcon = self.mediumIcon + 0.05
		self.wndMain:FindChild("txtMedium"):SetText((self.mediumIcon*100) .. "%")
	end
end

function Scryer:OnLargeDown( wndHandler, wndControl, eMouseButton )
	if self.largeIcon > 0 then
		self.largeIcon = self.largeIcon - 0.05
		self.wndMain:FindChild("txtLarge"):SetText((self.largeIcon*100) .. "%")
	end
end

function Scryer:OnLargeUp( wndHandler, wndControl, eMouseButton )
	if self.largeIcon < self.mediumIcon then
		self.largeIcon = self.largeIcon + 0.05
		self.wndMain:FindChild("txtLarge"):SetText((self.largeIcon*100) .. "%")
	end
end

-----------------------------------------------------------------------------------------------
-- Scryer Instance
-----------------------------------------------------------------------------------------------
local ScryerInst = Scryer:new()
ScryerInst:Init()
